/*       Function: MakeTables
      Description: Creates databases
           Author: Tad Frysinger
     Date created: 03-29-93
           Syntax: MakeTables( <aDbfStructs>[,<lForce>] ) --> lSuccess

        Arguments: aDbfStruct   Array whose elements are arrays with
																the following structure:
                                {DBFNAME, DBFFIELDS, DBFEXT}

																The DBFFIELDS element is an array with
																the structure expected by dbcreate()

                   lForce       TRUE if tables should be created even
                                though they exist

          Example:
     Return Value: TRUE if table was created, else FALSE
         See Also:
*/
#define DBFNAME       1
#define DBFFIELDS     2
#define DBFEXT        3
#define TRUNCATE_NIL  0
#define TRUNCATE_NONE 1
#define TRUNCATE_ONE  2
#define TRUNCATE_ALL  3
#define OLDSTRUCT     1
#define NEWSTRUCT     2
function MakeTables( aDbfStructs, lForce )
  local cScr              // Screen image
  local aDbf              // Current table info
  local aBar1             // Reference to total # of tables progress bar
  local aBar2             // Reference to records processed progress bar
  local oError            // Error object
  local i, x, y, z        // Generic counters
  local cName             // Current table name
  local aStruct           // Current table structure
  local lStructSame       // TRUE if new structure is defined as
                          // the same as the old structure, else FALSE
  local nTruncateData     // Holds the values defined as:
                          // TRUNCATE_NIL      No user intervention
                          // TRUNCATE_NONE     User said not to make changes
                          // TRUNCATE_ONE      User confirmed one field
                          // TRUNCATE_ALL      User said to accept any and
                          //                   all changes.  This is useful
                          //                   when several changes are
                          //                   made, so that the same warning
                          //                   message doesn't keep appearing.
  local lHasData          // There are records in the table
  local aFields           // New field structure
  local cDataPath         // Where the tables are located
  local cAlias            // Holds unique file name used to create new table
  local lRetval := FALSE  // If TRUE all tables were made
  local aDir              // Directory of files created.  This is used
                          // to make sure all files created are renamed
                          // to the old table name(s)
  local nDirFiles         // Number of files created (loop control)
  local aFieldMaps        // When we compare the old structure to the new,
                          // we need to know which field in the old to
                          // copy into the new.  This may not be all fields,
                          // as in the case where the structure changes
                          // such that a field no longer exists, of when
                          // a field is in a different ordinal position in
                          // the structure for the new table, as compared to
                          // the old table.  This array holds one element
                          // for each field in the old that should be copied
                          // to the new.  Each element is a two element
                          // array.  The first element holds the ordinal
                          // position in the old structure, and the second
                          // element holds the corresponding ordinal position
                          // in the new strucuture.
                          //
                          // For example, assume the field FIRSTNAME is in
                          // both structures, but it is the 2nd field in the
                          // old structure, and the 3rd field in the new
                          // structure,  The element in aFieldMaps that
                          // would handle transferring data would look like:
                          //
                          //  {2,3}
  local nOldArea          // Workarea for the old structure
  local nNewArea          // Workarea for the new structure
  local nOldReccount      // Record count for the old structure.  This is
                          // used to calculate the progress bar.
  local nFldCount         // Number of entries in aFieldMaps
  local nFldPos           // Used to determine whether the new structure
                          // contains fields not in the old structure.
  local nRecs             // Counter for records processed
  local aDict             // Contains a reference to the global data dict.
  local nArea             // Workarea when we started this function
  local lSysActive        // Whether or not the SYSTEM file is active

  begin sequence

  aDict      := CCMDict()
  cDataPath  := DictAt( aDict, "DATA_PATH" )
  cAlias     := DictAt( aDict, "TEMPFILE1")
  lSysActive := ( select("SYSTEM") > 0)
  default lForce to FALSE

  aBar1 := GaugeNew( 14,;
                     22,;
                     14,;
                     58,;
                     DictAt( aDict, "COLORMAIN")[MAIN_NORMAL],;
                     DictAt( aDict, "COLORMAIN")[MAIN_BOLD] )
  aBar2 := GaugeNew( 19,;
                     22,;
                     17,;
                     58,;
                     DictAt( aDict, "COLORMAIN")[MAIN_NORMAL],;
                     DictAt( aDict, "COLORMAIN")[MAIN_BOLD] )


  dispbegin()

  cScr := MakeBox( 10, 20, 22, 60, B_SINGLE + " ", TRUE )

  @ 11,35 say "Make Tables"
  @ 13,22 say "Tables:"
  @ 18,22 say "Progress:"
  GaugeDisplay( aBar1 )
  GaugeDisplay( aBar2 )

  dispend()

  if !LogOutUsers()

    break

  endif

  if (aDbfStructs == NIL)

    if !lSysActive

      oError := errornew()
      oError:genCode := ERR_SUBSYS_FAIL
      oError:description := "SYSTEM table unavailable"
      oError:filename := "SYSTEM.dbf"
      CALLSTACK IN oError
      break( oError )

    endif
    nArea := select()
    select SYSTEM
    // If we are using the SYSTEM file, then we have already installed
    locate for SYSTEM->Code == "TBLSTRCT"
    if !RecLock( USER_CAN_RETRY )

      select (nArea)
      break

    endif
    aDbfStructs := ARead( SYSTEM->Memo1 )
    dbunlock()
    select (nArea)

  endif

  // cycle through and create files
  x := len( aDbfStructs )
  aFieldMaps := array( x )

  // Add a fieldmap reference for each table
  for i := 1 to x

    aFieldMaps[ i ] := {}

  next

  for i := 1 to x

    // reset the gauge for creating the DBF
    GaugeUpdate( aBar2, 0 )

    aDbf  := aDbfStructs[i]
    cName := trim( aDbf[ DBFNAME ] )

    @ 13,32 say padr( cName, 8)

    if !lForce .and. ;
       file( cDataPath + cName + "." + aDbf[ DBFEXT ] )

      // If found, then we need to see if an alteration is
      // necessary
      use (cDataPath+cName) new shared alias (cAlias)
      aStruct := dbstruct()
      lHasData := reccount() > 0
      use

      // Get a separate copy of the new field structure.
      // We will use this to delete fields that are found, so
      // that we will know if there are any fields in the new
      // structure NOT in the old structure
      aFields  := aclone( aDbf[ DBFFIELDS ] )

      // Compare each field in the current table structure (held in
      // 'aStruct' with the new structure held in 'aFields'.  If there is
      // even one field with any different structure (type, length, dec)
      // then an alteration of the structure is necessary.  If the
      // alteration means that data could be lost, then alert the user
      nTruncateData := TRUNCATE_NIL  // No choices made
      lStructSame := TRUE

      for y := 1 to len( aStruct )

        if (z := ascan( aDbf[ DBFFIELDS ], ;
                {|aEle| aEle[DBS_NAME] == aStruct[y, DBS_NAME] } )) > 0

          // A field in the old structure is present in the new
          //
          // Add an entry that points from the old to the new.
          aadd( aFieldMaps[ i ], {y,z} )

          do case
          case lHasData .and. ;
               !( aDbf[ DBFFIELDS, z, DBS_TYPE ] == aStruct[y, DBS_TYPE] )
            // There are records present and the data type is different.
            // this means that data loss is possible

            lStructSame := FALSE

            if !(nTruncateData == TRUNCATE_ALL) .and. ;
               !((nTruncateData := ;
                  DataMsg( aDbf[DBFFIELDS, z, DBS_NAME], cName))>1 )

              // In this case, the user hasn't already indicated to
              // truncate all fields that are different, AND indicates
              // not to truncate *any*.

              nTruncateData := TRUNCATE_NONE
              lStructSame := TRUE

            else

              // User indicates that truncating data is ok.
              // Delete from the map file so data is not attempted to be
              // copied
              adel( aFieldMaps[ i ], len( aFieldMaps[i]))
              asize( aFieldMaps[ i ], len( aFieldMaps[ i ])-1)

            endif

          case !( aDbf[DBFFIELDS, z, DBS_LEN  ] == aStruct[y, DBS_LEN ] )

            lStructSame := FALSE
            // Find out if data truncation is a possibility
            if lHasData .and. ;
               aDbf[DBFFIELDS, z, DBS_LEN ] < aStruct[ y, DBS_LEN ]

              if !(nTruncateData == TRUNCATE_ALL) .and. ;
                 !((nTruncateData := ;
                    DataMsg( aDbf[DBFFIELDS, z, DBS_NAME], cName))>1 )

                // Set flag so no restructure will be performed
                nTruncateData := TRUNCATE_NONE
                lStructSame := TRUE
                // Make field structure the way it was
                aDbf[DBFFIELDS, z] := aStruct[ y ]

              endif

            endif

          case lHasData .and. ;
              !( aDbf[DBFFIELDS, z, DBS_DEC  ] == aStruct[y, DBS_DEC ] )
            // Data problems can occur if the decimal number moves up or
            // down, since the fixed length of the field includes the
            // decimal number.

            lStructSame := FALSE

            if !(nTruncateData == TRUNCATE_ALL) .and. ;
               !((nTruncateData := ;
                  DataMsg( aDbf[DBFFIELDS, z, DBS_NAME], cName))>1 )

              // Set flag so no restructure will be performed
              lStructSame := TRUE
              nTruncateData := TRUNCATE_NONE
              // Make field structure the way it was
              aDbf[DBFFIELDS, z] := aStruct[ y ]

            endif

          endcase

          // Get rid of the field in the aFields array that was found in
          // the current structure.  At the end of this cycle, if any
          // entries are left in 'aFields', they represent fields that
          // are in the new structure that weren't in the old.
          if ((nFldPos := ascan( aFields, ;
                            {|aEle| aEle[DBS_NAME] == ;
                                    aStruct[y, DBS_NAME] })) > 0)

            adel( aFields, nFldPos)
            asize( aFields, len( aFields)-1)

          endif

        else
          // A field in the old structure is NOT present in the new, which
          // means that loss of data will occur in that field for the
          // current structure
          lStructSame := FALSE
          if !(nTruncateData == TRUNCATE_ALL) .and. ;
             !((nTruncateData := ;
                DataMsg( aStruct[y,DBS_NAME], cName))>1 )

            // Set flag so no restructure will be performed
            lStructSame := TRUE
            nTruncateData := TRUNCATE_NONE

          endif

        endif

      next

      if len( aFields ) > 0

        // There is one or more new fields added to the existing
        // structure
        lStructSame := FALSE

      endif

      if !lStructSame

        // We need to create the new structure and append in the
        // old records.
        //
        // Open the old file
        use (cDataPath + cName) new exclusive
        nOldArea := select()
        nOldReccount := reccount()

        // Create a temp file with the new structure and open it
        dbcreate( cDataPath+cAlias, aDbf[ DBFFIELDS ])
        use (cDataPath + cAlias) new exclusive
        nNewArea := select()

        // Now append records from the old into the new
        @ 18,32 say "Appending:"


        select (nOldArea)
        go top
        nFldCount := len( aFieldMaps[ i ] )
        nRecs := 0
        while !eof()

          // Both are used exclusively
          (nNewArea)->( dbappend() )
          for nFldPos := 1 to nFldCount

            // For each field in the new structure that has a
            // corresponding field in the old structure, place
            // data from the old to the new
            (nNewArea)->( fieldput( aFieldMaps[i, nFldPos, NEWSTRUCT],;
            (nOldArea)->( fieldget( aFieldMaps[i, nFldPos, OLDSTRUCT]))))

          next
          nRecs++

          // Show the # records every 100
          if ( nRecs % 100 ) == 0

            @ 18,44 say transform( nRecs, "999,999,999,999" )

          endif

          // Update progress bar
          GaugeUpdate( aBar2, round( (nRecs / nOldreccount ), 2 ) )
          skip

        end
        (nNewArea)->( dbcommit() )
        (nNewArea)->( dbclosearea() )
        use

        // What files did we just make associated with the new table?
        aDir := directory( cDataPath + cAlias + ".*" )

        for nDirFiles := 1 to len( aDir )

          // Delete all files associated with the source file
          ferase( cDataPath + cName + ;
                  right( aDir[nDirFiles, F_NAME], 4 ) )

          // Rename from temp file with extension, to
          // old file with same extension as temp file
          frename( cDataPath + aDir[nDirFiles, F_NAME],;
                   cDataPath + cName + ;
                   right( aDir[nDirFiles, F_NAME],4))

        next

        @ 18,32 say space( 27 )

      endif

    else

      dbcreate( cDataPath + cName, aDbf[ DBFFIELDS ] )

    endif

    GaugeUpdate( aBar1, round( (i / x ), 2 ) )
    GaugeUpdate( aBar2, 1)


  next

  // Now get rid of the system message
  if !LogInUsers()

    oError := errornew()
    oError:genCode := ERR_SUBSYS_FAIL
    oError:description := "Couldn't lock SYSMSG for update"
    oError:filename := "SYSTEM.dbf"
    CALLSTACK IN oError
    break( oError )

  endif

  lRetval := TRUE

  PressKeyMsg({{"               Table Restructure Successful!           ",;
                " ",;
                " The table restructure has passed successfully.  You may",;
                " wish to reorder the tables at this time, if any of the",;
                " structural changes have rendered the keys inaccurate."}})

  recover using oError

    do case
    case oError:genCode == EG_CREATE .or.;
         oError:genCode == EG_CORRUPTION

      oError:genCode := ERR_FILE_CREATE

    case oError:genCode == EG_DATATYPE .or.;
         oError:genCode == EG_DATAWIDTH

      oError:genCode := ERR_DBF_STRUCT

    otherwise

      CALLSTACK IN oError

    endcase
    break( oError )

  end sequence

  RestBox( cScr )

  return( lRetval )
// eof MakeTables
#undef DBFNAME
#undef DBFFIELDS
#undef DBFEXT
#undef TRUNCATE_NIL
#undef TRUNCATE_NONE
#undef TRUNCATE_ONE
#undef TRUNCATE_ALL
#undef OLDSTRUCT
#undef NEWSTRUCT
